from pathlib import Path

import typer
from rich import print

from yudown import audio_dir, not_spec_dir, playlist_dir, video_dir


def VerifyLink(link: str):
    if "youtu" not in link:
        print("The link entered is not a valide Youtube link")
        raise typer.Abort()
    else:
        return link
    
def validate_choice(choice: str):
    valid_choice = ["audio", "video", "playlist"]
    if choice.lower() not in valid_choice:
        raise typer.BadParameter(f"{choice} is not a valid option. Valid option are {', '.join(valid for valid in valid_choice)}")
    else:
        return choice.lower()

def validate_location(location: Path):
    if not location.is_dir():
        raise typer.BadParameter(f"The path specified {location} is not a directory")
    return location

def choise_dir(location: Path, choice: str):
    if location == not_spec_dir:
        match choice:
            case "audio":
                location = audio_dir
            case "video":
                location = video_dir
            case "playlist":
                location = playlist_dir
    return location


import datetime

class Media:
    def __init__ (self, filename, extension, resolution, link, date_downloaded=None, id=None):
        self.filename = filename
        self.extension = extension
        self.resolution = resolution
        self.link = link
        self.date_downloaded = date_downloaded if date_downloaded is not None else datetime.datetime.now().isoformat()
        self.id = id

    def __repr__ (self) -> str:
        """Make the class more readeable
        """
        return f"({self.filename}, {self.extension}, {self.resolution}, {self.link}, {self.date_downloaded}, {self.id}"


"""This module provides the YuDownloader CLI."""
# yudown/cli.py

import time
from pathlib import Path
from typing import List, Optional

import typer
from pytube import Search, YouTube, exceptions
from pytube.cli import on_progress
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Prompt
from rich.table import Table

from yudown import __app_name__, __version__, not_spec_dir, playlist_dir
from yudown.database import create, destroy, read
from yudown.model import Media
from yudown.playlist import PlaylistObject
from yudown.utils import (VerifyLink, choise_dir, validate_choice,
                          validate_location)

app = typer.Typer(rich_markup_mode='rich')


def _version_callback(value: bool) -> None:
    if value:
        print(f"[bold red]{__app_name__}[/bold red] [green]v{__version__}[/green]")
        raise typer.Exit()


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the application's version and exit.",
        callback=_version_callback,
        is_eager=True,
    ),
) -> None:
    return


@app.command()
def history(
    clear: Optional[bool] = typer.Option(
        None,
        "--delete",
        "-D",
        show_default=False,
    )
    ):
    """ Show the download [blue]history[/blue] ⌚️"""
    
    if clear:
        delete = typer.confirm("Are you sure to clear all history ?", abort=True)
        if delete:
            print("History deleted")
            destroy()
            raise typer.Exit()
            
    media = read()

    print("📜", "[bold magenta]Download History[/bold magenta]", "📜")

    if len(media) == 0:
        print("[bold red]No download history to show[/bold red]")
    else:
        table = Table(show_header=True,
                      header_style="bold", show_lines=True)
        table.add_column("#", style="dim", width=3, justify="center")
        table.add_column("Filename", min_width=20, justify="center")
        table.add_column("Extension", min_width=20, justify="center")
        table.add_column("Resolution", min_width=15, justify="center")
        table.add_column("Link", min_width=30, justify="center")

        for idx, media in enumerate(media, start=1):
            table.add_row(
                str(idx), f'[cyan]{media.filename}[/cyan]', f'[yellow]{media.extension}[/yellow]', f'[green]{media.resolution}[/green]', f'[red]{media.link}[/red]')
        print(table)
        
        
@app.command("search")
def make_search(
    search_query: Optional[str] = typer.Argument("", help="The word you are searching for", show_default=False),
    suggestion: Optional[bool] = typer.Option(
        None,
        "--suggestion",
        "-s",
        help="Show search suggestion",
        show_default=False
    )
    ):
    """[blue]Search[/blue] for video on Youtube 🔍"""
    
    if search_query == "":
        search_query = Prompt.ask("Enter a word to search for: ")
    
    s = Search(search_query)
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=False,
    ) as progress:
        progress.add_task(description="Searching...", total=None)
        time.sleep(60)
    
    table = Table(
        show_header=True,
        header_style="bold",
        show_lines=True
    )
    table.add_column("#", style="dim", width=3, justify="center")
    table.add_column("Title", min_width=30, justify="center")
    table.add_column("Link", min_width=20, justify="center")
    
    resultat = s.results
    for idx, resultat in enumerate(resultat, start=1):
        table.add_row(
                str(idx), f'[cyan]{resultat.title}[/cyan]', f'[yellow]{resultat.watch_url}[/yellow]')
    print(table)
    
    if suggestion:
        for i in s.completion_suggestions:
            print(f"{i}")
        print(f"[bold green]{len(s.completion_suggestions)}[/bold green] suggestions founded\n")
    
    raise typer.Exit()


@app.command("playlist")
def PlaylistDownload(
    link: Optional[str] = typer.Argument(
        None,
        show_default=False
    ),
    locate: Optional[Path] = typer.Option(
        playlist_dir,
        "--location",
        "-l",
        help="Location of the files to download",
        show_default=True
    )
):
    """Download Youtube [yellow]Playlist[/yellow] video 📼"""
    if link == None:
        #link of the video to be downloader
        link = input("Enter the playlist link")
    
    url = VerifyLink(link)
    playlist = PlaylistObject(url)
    
    try:
        print(f"Downloading {playlist.title}")
        for video in playlist.videos:
            table = Table(
        show_header=True,
        header_style="bold",
        show_lines=True
    )
            table.add_column("#", style="dim", width=3, justify="center")
            table.add_column("Title", min_width=30, justify="center")
            table.add_column("Link", min_width=20, justify="center")
            
            for idx, video in enumerate(video, start=1):
                table.add_row(
                        str(idx), f'[cyan]{video.title}[/cyan]', f'[yellow]{video.watch_url}[/yellow]')
            print(table)
            
            down = typer.confirm("Download this playlist ?")
            if not down:
                raise typer.Abort()
            video.streams.first().download(locate)
    except:
        print("Some error, download not completed !")
        raise typer.Abort()
    else:
        play = Media(filename=video.title, extension="mp4", resolution="Playlist", link=video.watch_url)
        create(play)
        
        print("Downloaded successfully !")
        raise typer.Exit()


@app.command()
def download(    
    type: Optional[str] = typer.Option(
        "video",
        "--type",
        "-t",
        help="The type of media to download",
        show_default=True,
        callback=validate_choice,
    ),
    location: Optional[Path] = typer.Option(
        not_spec_dir,
        "--location",
        "-l",
        help="Location of the downloaded file",
        show_default=True,
        callback=validate_location,
    ),
    links: Optional[List[str]] = typer.Argument(None, show_default=False)
):
    """Download file from [red]Youtube[/red] 📥"""
    if not links:
        links.append(VerifyLink(input("Please enter link to download")))
    
    for url in links:
        try:
			# object creation using YouTube
			# which was imported in the beginning
            link = VerifyLink(url)
            yt = YouTube(link, on_progress_callback=on_progress)
            title = yt.title
            
        except exceptions.VideoPrivate:
            print("[bold red]Media is private ![/bold red]") #to handle exception
            
        except exceptions.VideoRegionBlocked:
            print("[bold red]Media is blocked ![/bold red]")
            
        except exceptions.VideoUnavailable:
            print("[bold red]Media is not available ![/bold red]")
            
        else:
            match type:
                case "audio":
                    audiobj = yt.streams.order_by('mime_type').filter(type='audio')
                    
                    fileExtension = [stream.mime_type for stream in audiobj]
                    fileAudio = [stream.audio_codec for stream in audiobj]
                    audio = [stream for stream in audiobj]
                    
                    locate = choise_dir(location, "audio")

                    i = 1
                    
                    for extension in fileExtension:
                        print(f'👉 {i}- Extension: [yellow]{extension}[/yellow] -> Audio: [bold green]{fileAudio[i-1]}[/bold green]')
                        i += 1
                        
                    # To Download the video with the users Choice of resolution
                    strm = int(input('\nChoose a resolution please: '))
                                    
                    if 1 <= strm < i:
                        try:
                            # To validate if the user enters a number displayed on the screen...
                                extension_to_download = fileExtension[strm-1]
                                print(f"You're now downloading the audio with extension {extension_to_download}...")
                                # command for downloading the video
                                file = audio[strm-1]
                                file.download(locate)
                        except:
                            print("❌[bold red]Some Error, download not completed ![/bold red]❌")
                            raise typer.Abort()
                        else:
                            print("\n[bold green]Downloaded successfully ![/bold green] 🥳")
                            media = Media(filename=title, extension="audio", resolution=fileExtension[strm-1], link=url)
                            create(media)
                            raise typer.Exit()
                    else:
                        print("[red]Invalid choice !![/red]\n\n")
                        raise typer.Abort()
                
                case "video":                    
                    locate = choise_dir(location, "video")
                    dwn = yt.streams.filter(type="video", progressive="True", file_extension='mp4')
                    print("\n👉 1- [bold green]Highest[/bold green] resolution\n👉 2- [bold red]Lowest[/bold red] resolution\n👉 3- See all available")
                    res = int(input("Choose resolution: "))
                    print()
        
                    if res == 1:
                        try:
                            high = dwn.get_highest_resolution().resolution
                            
                            print(f"Downloading {title} [bold italic green]{high}[/bold italic green]")
                            
                            dwn.get_highest_resolution().download(locate, filename=title+" "+high)
                        except:
                            print(f"Error while downloading the video !")
                            raise typer.Abort()
                        else:
                            print("\nDownload success !!")
                            video1 = Media(filename=title, extension="mp4", resolution=high, link=url)
                            create(video1)
                            raise typer.Exit()
                        
                    elif res == 2:
                        try:
                            low = dwn.get_lowest_resolution().resolution
                            
                            print(f"Downloading {title} [bold italic red]{low}[/bold italic red]")
                            
                            dwn.get_lowest_resolution().download(locate, filename=title+" "+low)
                        except:
                            print(f"Error while downloading the video !")
                            raise typer.Abort()
                        else:
                            print("\nDownload success !!")
                            video2 = Media(filename=title, extension="mp4", resolution=low, link=url)
                            create(video2)
                            raise typer.Exit()

                    elif res == 3:
                        videobj = yt.streams.order_by('resolution').filter(progressive="True")
                        
                        video_resolutions = [stream.resolution for stream in videobj]
                        fileExtension = [stream.mime_type for stream in videobj]
                        videos = [stream for stream in videobj]		
                        i = 1
                        
                        for resolution in video_resolutions:
                            print(f'👉 {i}- [green]{resolution}[/green] -> Extension: {fileExtension[i-1]}')
                            i += 1

                        # To Download the video with the users Choice of resolution
                        strm = int(input('\nChoose a resolution please: '))
                        
                        if 1 <= strm < i:
                            try:
                                # To validate if the user enters a number displayed on the screen...
                                resolution_to_download = video_resolutions[strm-1]
                                
                                print(f"You're now downloading the video with resolution [bold italic green]{resolution_to_download}[/bold italic green]...")

                                videos[strm-1].download(locate, filename=title+" "+resolution_to_download)
                                
                            except:
                                print("[bold red]Some Error, download not completed ![/bold red]")
                                raise typer.Abort()
                            else:
                                print("\nDownloaded successfully !")
                                video3 = Media(filename=title, extension="mp4", resolution=resolution_to_download, link=url)
                                create(video3)
                                raise typer.Exit()
                        else:
                            print("Invalid choice !!\n\n")
                            raise typer.Abort()
    
                    else:
                        print("Error ! Enter a valid number !!")
                        raise typer.Abort()
                case "playlist":
                    locate = choise_dir(location, "playlist")
                    for i in links:
                        PlaylistDownload(i, locate)


"""Top-level package for YuDownloader."""
# yudownloader/__init__.py
from pathlib import Path

from tinydb import Query, TinyDB

__app_name__ = "YuDown"
__version__ = "0.4.0"

yudown_dir = Path.home()/"YuDown"
video_dir = yudown_dir/"Video"
audio_dir = yudown_dir/"Audio"
playlist_dir = yudown_dir/"Playlist"
not_spec_dir = yudown_dir/"notSpecified"

if not Path.exists(yudown_dir):
    Path(yudown_dir).mkdir(exist_ok=True)
    
if not Path.exists(video_dir):
    Path(video_dir).mkdir(exist_ok=True)
    
if not Path.exists(audio_dir):
    Path(audio_dir).mkdir(exist_ok=True)
    
if not Path.exists(playlist_dir):
    Path(playlist_dir).mkdir(exist_ok=True)
    
if not Path.exists(not_spec_dir):
    Path(not_spec_dir).mkdir(exist_ok=True)

db = TinyDB(yudown_dir/"history.json")
db.default_table_name = 'media-history'
MediaQuery = Query()


from typing import Optional

import typer
from pytube import Playlist, exceptions
from rich import print


def PlaylistObject(link: Optional[str]):    
    try:
        # object creating using Youtube
        pylst = Playlist(link)
    except exceptions.VideoPrivate:
        print("[bold red]Media is private ![/bold red]")
        raise typer.Abort()
    except exceptions.VideoRegionBlocked:
        print("[bold red]Media is blocked ![/bold red]")
        raise typer.Abort()
    except exceptions.VideoUnavailable:
        print("[bold red]Media is not available ![/bold red]")
        raise typer.Abort()
    else:
        return pylst


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yudown']

package_data = \
{'': ['*']}

install_requires = \
['pytube>=12.1.2,<13.0.0', 'tinydb>=4.7.1,<5.0.0', 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['yudown = yudown.cli:app']}

setup_kwargs = {
    'name': 'yudown',
    'version': '0.4.0',
    'description': 'This is a python script to download Youtube Media file',
    'long_description': '# YuDown (project under improvement)\n\nYuDown is a python script to download YouTube video developped using [Typer](https://typer.tiangolo.com) and [Pytube](https://github.com/pytube/pytube)\n\n## Installation\n\nTo install and run the script:\n\n```bash\n  python3 -m venv venv\n  pip install yudown\n\n  yudown --help\n```\n\n## Run Locally\n\nTo launch the project, you need [Poetry](https://python-poetry.org) to be installed\n\nClone the project\n\n```bash\n  git clone https://github.com/TianaNanta/yudown.git\n```\n\nGo to the project directory\n\n```bash\n  cd yudown\n```\n\nInstall dependencies\n\n```bash\n  poetry install\n```\n\nLaunch the project\n\n```bash\n  poetry run python -m yudown --help\n```\n\n## Usage/Examples\n\nTo download audio file from the given Youtube link\n\n```bash\n  yudown -A https://youtube.com/.....\n```\n\n## Authors\n\n- [@TianaNanta](https://www.github.com/TianaNanta)\n\n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'TianaNanta',
    'author_email': 'andrianjakananta@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List
from yudown.model import Media
from yudown import db, MediaQuery

def change_id(old_id: int, new_id: int) -> None:
    db.update({'id': new_id},
              MediaQuery.id == old_id)

def create(media: Media) -> None:
    count = len(db)            
    media.id = len(db)+1
    new_media = {
        'filename': media.filename,
        'extension': media.extension,
        'resolution': media.resolution,
        'link': media.link,
        'date_downloaded': media.date_downloaded,
        'id': media.id
    }
    db.insert(new_media)
    
    if count > 10:
        db.remove(MediaQuery.id == 1)
        for pos in range(2, count):
            change_id(pos, pos-1)

def read() -> List[Media]:
    results = db.all()
    media = []
    for result in results:
        new_media = Media(result['filename'], result['extension'], result['resolution'],
                              result['link'], result['date_downloaded'], result['id'])
        media.append(new_media)
    return media

def destroy() -> None:
    db.drop_tables()